import 'dart:io';

/// Base options class defining a Redis connection.
class RedisOptions {
  RedisOptions({
    this.host = 'localhost',
    this.port = 6379,
    this.database = 0,
    this.connectTimeout = const Duration(seconds: 10),
    this.password,
    this.username,
    this.name,
    this.tls = false,
    this.protocol = 2,
    this.socketOptions = const [],
  });

  /// Parse a Redis connection string protocol.
  /// See links for more information on URI schemes;
  ///  - redis: https://www.iana.org/assignments/uri-schemes/prov/redis
  ///  - rediss: https://www.iana.org/assignments/uri-schemes/prov/rediss
  factory RedisOptions.fromUri(Uri uri) {
    // Must be 'redis' or 'rediss' (TLS).
    if (uri.scheme != 'redis' && uri.scheme != 'rediss') {
      throw ArgumentError.value(
        uri.toString(),
        'url',
        'Invalid URL scheme "${uri.scheme}" - must be "redis" or "rediss".',
      );
    }

    // If absent, the default value of the "host" URI field is:
    // "localhost" (or equivalent)
    final host = uri.host.isEmpty ? 'localhost' : uri.host;

    // If absent, the default value of the "port" URI field is:
    // 6379.
    final port = uri.port == 0 ? 6379 : uri.port;

    // The database number to use for the Redis SELECT command comes from
    // either the "db-number" portion of the URI or the value from the key-value
    // pair from the "query" URI field with the key "db". If neither of these
    // are present, the default database number is 0.
    var database = 0;
    if (uri.pathSegments.isNotEmpty) {
      database = int.parse(uri.pathSegments.first);
    }
    if (uri.pathSegments.isEmpty && uri.queryParameters.containsKey('db')) {
      database = int.parse(uri.queryParameters['db']!);
    }

    // The password to use for the Redis AUTH command comes from either the
    // password portion of the "userinfo" URI field or the value from the
    // key-value pair from the "query" URI field with the key "password".
    // If neither of these are present, the client ought not to issue an
    // initial AUTH command.
    String? password;
    if (uri.userInfo.isNotEmpty) {
      final userInfo = uri.userInfo.split(':');
      password = userInfo.length > 1 ? userInfo[1] : null;
    }
    if ((password == null || password.isEmpty) &&
        uri.queryParameters.containsKey('password')) {
      password = uri.queryParameters['password'];
    }
    if (password != null && password.isEmpty) {
      password = null;
    }

    // For Redis ACL support, it's suggested that the
    // username to use when authenticating be obtained from the username
    // portion of the "userinfo" URI field.
    // We optional also support fallback to "username" from query parameters.
    String? username;
    if (uri.userInfo.isNotEmpty) {
      final userInfo = uri.userInfo.split(':');
      username = userInfo.isNotEmpty
          ? userInfo[0]
          : uri.userInfo.isNotEmpty
              ? uri.userInfo
              : null;
    }
    if ((username == null || username.isEmpty) &&
        uri.queryParameters.containsKey('username')) {
      username = uri.queryParameters['username'];
      if (username!.isEmpty) {
        username = null;
      }
    }

    return RedisOptions(
      host: host,
      port: port,
      database: database,
      password: password,
      username: username,
      tls: uri.scheme == 'rediss',
    );
  }

  /// Create a copy of a [RedisOptions] instance and optionally
  /// change it's values.
  factory RedisOptions.copy(
    RedisOptions options, {
    String? host,
    int? port,
    int? database,
    Duration? connectTimeout,
    String? password,
    String? username,
    String? name,
    int? protocol,
    bool? tls,
    List<SocketOption>? socketOptions,
  }) {
    return RedisOptions(
      host: host ?? options.host,
      port: port ?? options.port,
      database: database ?? options.database,
      connectTimeout: connectTimeout ?? options.connectTimeout,
      password: password,
      username: username,
      name: name,
      protocol: protocol ?? options.protocol,
      tls: tls ?? options.tls,
      socketOptions: socketOptions ?? options.socketOptions,
    );
  }

  /// The host of the Redis server.
  ///
  /// Defaults to `localhost`.
  final String host;

  /// The port of the Redis server.
  ///
  /// Defaults to `6379`.
  final int port;

  /// The Redis database to use.
  ///
  /// Defaults to `0`.
  final int database;

  /// Password to use when connecting to the Redis Server.
  /// When set, the connection will be authenticated after connecting to
  /// the server via the `AUTH` Redis command.
  final String? password;

  /// Username to use when connecting to the Redis Server
  /// for Redis ACL support.
  final String? username;

  /// Optional name of this client to set when connecting to Redis
  /// via the `CLIENT SETNAME` Redis command.
  final String? name;

  /// The timeout for connecting to the Redis server.
  ///
  /// Defaults to `10` seconds.
  final Duration connectTimeout;

  /// The Redis protocol version to use.
  ///
  /// Defaults to `2` and currently only `2` is supported.
  ///
  /// See: https://redis.io/topics/protocol
  final int protocol;

  /// Whether to use TLS when connecting to the Redis server.
  /// Note that Redis itself does not support TLS natively, so this option is
  /// dependent on your server hosting provider and if they support TLS,
  /// e.g. `redis.com` or if you are hosting behind a TLS proxy yourself.
  ///
  /// Defaults to `false`.
  final bool tls;

  /// Socket options to set as enabled when creating the socket connection
  /// to the Redis server.
  ///
  /// See [SocketOption] for more information.
  ///
  /// Defaults to `[]`.
  final List<SocketOption> socketOptions;
}
