import 'dart:async';

/// Commands that are usable whilst the Redis client is in subscriber mode.
// TODO can this be discovered from flags?
const List<String> subscriberModeAvailableCommands = [
  'SUBSCRIBE',
  'SSUBSCRIBE',
  'SUNSUBSCRIBE',
  'PSUBSCRIBE',
  'UNSUBSCRIBE',
  'PUNSUBSCRIBE',
  'PING',
  'RESET',
  'QUIT',
];

/// A representation of connection changes created by the `RedisConnector` when
/// the connection state changes.
class RedisConnectionStateChange {
  RedisConnectionStateChange({
    required this.currentState,
    required this.previousState,
  });

  /// The new state of the connection.
  final RedisConnectionState currentState;

  /// The previous state of the connection.
  final RedisConnectionState previousState;

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is RedisConnectionStateChange &&
          currentState == other.currentState &&
          previousState == other.previousState;

  @override
  int get hashCode => currentState.hashCode & previousState.hashCode;
}

/// A command that has been queued for execution on the Redis connector instance.
class QueuedRedisCommand {
  QueuedRedisCommand({
    required this.completer,
    required this.command,
    required this.stacktrace,
    this.args,
  });

  /// The completer that will be completed when the command has been executed.
  final Completer<Object?> completer;

  /// The command that is being executed.
  final String command;

  /// The arguments that are being passed to the command.
  final List<Object?>? args;

  /// The originating stacktrace of the command that is being executed.
  /// This helps provide better stack locations to end users that are sending
  /// commands, since the stacktrace provides the locati
  final StackTrace stacktrace;
}

/// State of the Redis client.
enum RedisConnectionState {
  /// The connector is attempting to connect to the Redis server.
  connecting,

  /// The connector is connected to the Redis server.
  ready,

  /// The connector is attempting to disconnect from the Redis server.
  /// This is usually only indicated when close() is called on the underlying
  /// connector.
  closing,

  /// The connector is disconnected from the Redis server.
  /// This is the default state for new instances of the connectors.
  closed,

  /// An error has occurred when attempting to connect to the Redis server
  /// or when the connnection is closed due to an error.
  /// Sent after the [closed] state.
  error,
}

/// Error message for commands that are not usable whilst the Redis client is
/// in the specified mode.
String errorCommandNotSupportedInMode(String command, String mode) {
  return 'Command "$command" is not supported in $mode mode.';
}

/// A representation of a Redis error when a command fails, e.g. when the
/// args are invalid or the node is not ready.
class RedisCommandError implements Error {
  RedisCommandError(
    this.message, {
    required this.command,
    this.args = const <String>[],
    required this.stackTrace,
  });

  /// The command that caused the error.
  final String command;

  /// The arguments passed to the command.
  final List<Object?>? args;

  /// The error message returned by Redis.
  final String message;

  @override
  final StackTrace stackTrace;

  @override
  String toString() =>
      'RedisCommandError: ${message.replaceFirst('ERR ', '')}\n  COMMAND: "$command"\n  ARGS: $args';
}

/// A representation of a Redis reply.
class RedisReply<T extends Object?> {
  RedisReply(
    this.value, {
    this.attributes,
  });

  /// The value of the reply.
  final T value;

  /// The attrubutes of the reply returned as additional information.
  /// This only applies to RESP 3.
  final Map<String, Object?>? attributes;

  @override
  String toString() => 'RedisReply: $value, attributes: $attributes';
}
