import 'dart:async';
import 'dart:io';
import 'dart:math';

import 'package:dart_redis_connector/dart_redis_connector.dart';
import 'package:spec/spec.dart';

void main() {
  group('RedisConnector', () {
    test('should timeout connecting if timeout option specified', () async {
      final redisOptions = RedisOptions(
        host: '1.1.1.1',
        port: 9999,
        connectTimeout: const Duration(seconds: 1),
      );
      final redis = RedisConnector(redisOptions);
      final connectFuture = redis.connect();
      await expect(connectFuture).throws.isA<SocketException>();
      // TODO validate message contains 'Connection timed out'
    });

    test('.state should reflect the current connection state of the connector',
        () async {
      final redisOptions = RedisOptions();
      final redis = RedisConnector(redisOptions);
      expect(redis.state).toBe(RedisConnectionState.closed);
      final connectingFuture = redis.connect();
      expect(redis.state).toBe(RedisConnectionState.connecting);
      await connectingFuture;
      expect(redis.state).toBe(RedisConnectionState.ready);
      final closingFuture = redis.close();
      expect(redis.state).toBe(RedisConnectionState.closing);
      await closingFuture;
      expect(redis.state).toBe(RedisConnectionState.closed);
    });

    test('built-in StreamView should emit [$RedisConnectionStateChange] events',
        () async {
      final redisOptions = RedisOptions();
      final redis = RedisConnector(redisOptions);
      final changeEvents = <RedisConnectionStateChange>[];
      final streamSubscription = redis.listen(changeEvents.add);
      await redis.connect();
      await redis.close();
      await streamSubscription.cancel();

      expect(changeEvents).toEqual([
        RedisConnectionStateChange(
          previousState: RedisConnectionState.closed,
          currentState: RedisConnectionState.connecting,
        ),
        RedisConnectionStateChange(
          previousState: RedisConnectionState.connecting,
          currentState: RedisConnectionState.ready,
        ),
        RedisConnectionStateChange(
          previousState: RedisConnectionState.ready,
          currentState: RedisConnectionState.closing,
        ),
        RedisConnectionStateChange(
          previousState: RedisConnectionState.closing,
          currentState: RedisConnectionState.closed,
        ),
      ]);
    });

    test(
        'calling send should add the command to the [commandQueue] and remove when replied',
        () async {
      final redisOptions = RedisOptions();
      final redis = RedisConnector(redisOptions);
      await redis.connect();
      final commandFuture = redis.send<String>('SET', args: ['key', 'value']);
      expect(redis.commandQueue.length).toEqual(1);
      final redisReply = await commandFuture;
      expect(redis.commandQueue.length).toEqual(0);
      expect(redisReply.value).toEqual('OK');
      await redis.close();
    });

    test('calling connect with an unsupported protocol throws', () async {
      final redisOptions = RedisOptions(protocol: 3);
      final redis = RedisConnector(redisOptions);
      final connectFuture = redis.connect();
      await expect(connectFuture).throws.isA<UnsupportedError>();
      await redis.close();
    });

    test('calling connect returns and instance of $RedisConnectionInfo',
        () async {
      final redisOptions = RedisOptions();
      final redis = RedisConnector(redisOptions);
      final connectionInfo = await redis.connect();
      expect(connectionInfo.protocol).toEqual(redisOptions.protocol);
      expect(connectionInfo.mode).toEqual('standalone');
      await redis.close();
    });

    test(
        'calling closeWithError should send an error state to the stream and set the state to RedisConnectionState.error',
        () async {
      const errorMessage = 'WOOPS';
      final redisOptions = RedisOptions();
      final redis = RedisConnector(redisOptions);
      final changeEvents = <RedisConnectionStateChange>[];
      final errors = <String>[];
      final streamSubscription =
          redis.listen(changeEvents.add, onError: errors.add);
      await redis.connect();
      await redis.closeWithError(
        errorMessage,
        StackTrace.current,
      );
      expect(redis.state).toEqual(RedisConnectionState.error);
      expect(changeEvents.last).toEqual(
        RedisConnectionStateChange(
          previousState: RedisConnectionState.closed,
          currentState: RedisConnectionState.error,
        ),
      );
      expect(errors).toEqual([errorMessage]);
      await streamSubscription.cancel();
    });

    test('should allow subscribing to redis pub sub messages', () async {
      final redisOptions = RedisOptions();
      final redisSubscriber = RedisConnector(redisOptions);
      final redisPublisher = RedisConnector(redisOptions);
      await redisSubscriber.connect();
      await redisPublisher.connect();
      const channel = 'test_channel';
      const message = 'test_message';
      final stream = redisSubscriber.subscribeToMessages(
        channelOrPattern: channel,
        subscribeCommand: 'SUBSCRIBE',
        unsubscribeCommand: 'UNSUBSCRIBE',
      );
      final streamFirstFuture = stream.first;
      await Future<void>.delayed(Duration.zero);
      await redisPublisher.send('PUBLISH', args: [channel, message]);
      await Future<void>.delayed(Duration.zero);
      final event = await streamFirstFuture;
      expect(event[1]).toEqual(channel);
      expect(event[2]).toEqual(message);
      await redisSubscriber.close();
      await redisPublisher.close();
    });

    test(
        'should close the subscribeToMessages streams when the redis connector itself is closed',
        () async {
      final redisOptions = RedisOptions();
      final redisSubscriber = RedisConnector(redisOptions);
      await redisSubscriber.connect();
      const channel = 'test_channel';
      final onDoneCompleter = Completer<void>();
      redisSubscriber
          .subscribeToMessages(
            channelOrPattern: channel,
            subscribeCommand: 'SUBSCRIBE',
            unsubscribeCommand: 'UNSUBSCRIBE',
          )
          .listen(null, onDone: onDoneCompleter.complete);
      await Future<void>.delayed(Duration.zero);
      await redisSubscriber.close();
      await onDoneCompleter.future;
    });

    test('should scan keys matching a pattern', () async {
      final redisOptions = RedisOptions();
      final redis = RedisConnector(redisOptions);
      await redis.connect();
      final random = Random();
      final keys = <String>[];
      for (var i = 0; i < 1000; i++) {
        final key = 'key_$i';
        final value = random.nextInt(20000);
        await redis.send('SET', args: [key, value]);
        keys.add(key);
      }

      final scanAllResult = await redis.scan('SCAN', match: 'key_*').toList();
      expect(scanAllResult.length).toEqual(1000);
      for (var i = 0; i < 1000; i++) {
        expect(scanAllResult.contains(keys[i])).toEqual(true);
      }

      final scanJustOne = await redis.scan('SCAN', match: 'key_999*').toList();
      expect(scanJustOne.length).toEqual(1);
      expect(scanJustOne[0]).toEqual('key_999');

      await redis.close();
    });
  });
}
